"""
This module is imported on app ready (see __init__).
"""

from django.db.models.signals import m2m_changed
from django.dispatch import receiver
from .tools import models, relation

@receiver(m2m_changed, sender=models.ALL['Association'].associations.through)
def validate_associations(**kwargs):
    "Validate associations."
    if kwargs['action'] != 'pre_add':
        return

    if kwargs['reverse']:
        return

    ids = kwargs['pk_set']
    stems = relation.get_stems_by_id(kwargs['instance'].organisation.id,
                                     recursive=True)

    existing = set()
    for row in relation.flatten(stems):
        for entry in row:
            existing.add(entry)

    for entry in ids:
        if entry in existing:
            text = "Circular associations error on adding '%s'."
            text = text % relation.get_name(entry)
            raise ValueError(text)



"""
Fetching models via app.
"""
from django.apps import apps as _apps
from .. import __info__

# pylint: disable=no-member, protected-access
# We go through this effort so that we can use common everywhere and don't have
# to be concerned about import conflicts.
ALL = dict()
for _ in _apps.get_app_config(__info__.LABELS['name']).get_models():
    if _._meta.app_label == __info__.LABELS['name']:
        ALL[_.__name__] = _
        globals()[_.__name__] = _





"""
We just simply add all models in the admin interface.
"""
from django.contrib import admin
from django.contrib.admin.sites import AlreadyRegistered
from .tools import models

# First define custom admin on required models.

# Now register the remaining with the default settings.
for key in models.ALL:
    subject = models.ALL[key]

    try:
        admin.site.register(subject)
    except AlreadyRegistered:
        pass






"""
The Abstracted base for the models.
"""
from django.db import models
from django.utils.timezone import now
# pylint: disable=too-few-public-methods, no-member

class QueryManager(models.Manager):
    "Query Manager to restrict returning data"
    def get_queryset(self):
        "get_queryset override"
        query = models.Manager.get_queryset(self)
        query.filter(dts_delete__isnull=False)
        return query


class Abstract(models.Model):
    "Abstract share by all subsequent models."
    objects = QueryManager()

    class Meta:
        "Meta section to identify this is abstract"
        abstract = True

    dts_insert = models.DateTimeField(auto_now_add=True)
    dts_update = models.DateTimeField(null=True, blank=True)
    dts_delete = models.DateTimeField(null=True, blank=True)

    def save(self, force_insert=False, force_update=False, using=None,
        update_fields=None):
        if self.id not in [0, None]:
            self.dts_update = now()
        returns = super().save(force_insert=force_insert,
                               force_update=force_update,
                               using=using, update_fields=update_fields)
        return returns

    def __str__(self):
        text = str(self.id)
        if hasattr(self, 'label'):
            text = str(self.label)

        return text



"""
Tools for handling the relations.
"""
from collections import defaultdict
from functools import lru_cache
from . import models

# pylint: disable=no-member, redefined-variable-type

@lru_cache()
def lookups():
    "Return lookups"
    tmp = dict()
    _ = models.Organisation.objects.all().values_list('id', 'label')
    tmp['name'] = dict(_)
    tmp['name'][0] = '∞'
    tmp['id'] = {tmp['name'][key]: key for key in tmp['name']}
    tmp['stem_leafs'] = defaultdict(set)
    tmp['leaf_stems'] = defaultdict(set)
    tmp['stems'] = set()
    tmp['leafs'] = set()
    _ =  models.Association.objects.all().values_list('organisation__id',
                                                      'associations__id')
    for stem, leaf in _:
        tmp['stems'].add(stem)
        tmp['leafs'].add(leaf)
        tmp['stem_leafs'][stem].add(leaf)
        tmp['leaf_stems'][leaf].add(stem)

    tmp['head'] = tmp['stems'].difference(tmp['leafs'])
    tmp['tail'] = tmp['leafs'].difference(tmp['stems'])
    tmp['node'] = tmp['stems'].intersection(tmp['leafs'])

    return tmp


def _id_to_name_recursive(container):
    "converts a list of ids to a list of names"
    names = lookups()['name']
    if isinstance(container, (int,)):
        return names[container]

    if isinstance(container, dict):
        _ = dict()
        for key, value in container.items():
            key = names[key]
            _[key] = _id_to_name_recursive(value)
        return _

    if isinstance(container, (set, tuple, list)):
        _ = list()
        for entry in container:
            _.append(_id_to_name_recursive(entry))
        if len(_) == 1:
            return _[0]
        else:
            return _

    text = 'Unexpected container type %s with value %s'
    text = text % (str(type(container)), str(container))
    raise ValueError(text)


def get_heads_as_ids():
    "Return all organisations that have no parent association."
    tmp = lookups()
    return tmp['head']

def get_heads():
    "Same as get_heads_ids but lookup names."
    return _id_to_name_recursive(get_heads_as_ids())

def get_tails_as_ids():
    "Return all organisations that have no parent association."
    tmp = lookups()
    return tmp['tail']

def get_tails():
    "Same as get_tails_ids but lookup names."
    return _id_to_name_recursive(get_tails_as_ids())

def get_nodes_as_ids():
    "Return organisations that are neither head nor tail."
    tmp = lookups()
    return tmp['node']

def get_nodes():
    "Same as get_roots_ids but lookup name."
    return _id_to_name_recursive(get_nodes_as_ids())

def get_name(identifier):
    "Return the name by the id."
    tmp = lookups()
    return tmp['name'][identifier]

def get_id(name):
    "Return the id by the name."
    tmp = lookups()
    return tmp['id'][name]

def _get_relations_by_id(key, idx, recursive=False, root=None):
    """Get relations, key is either 'tail' or 'head'.
    This means that if tail is specified the idx is followed to all its tails.
    Alternatively if head is specified the idx is followed to all its heads.
    """
    tmp = lookups()
    if key == 'tail':
        related = tmp['stem_leafs'][idx]
    elif key == 'head':
        related = tmp['leaf_stems'][idx]
    else:
        raise ValueError("Key must be either 'tail' or 'head', not %s" % key)

    related = list(related)
    related.sort()

    if not recursive:
        return related

    returns = {idx:list()}

    if root is None:
        root = [idx]

    for entry in related:
        if entry in tmp[key]:
            append = entry
        else:
            if entry in root:
                append = {entry: 0}
            else:
                root.append(entry)
                append = _get_relations_by_id(key, entry, recursive, root)

        returns[idx].append(append)

    return returns

def get_leafs_by_id(organisation_id, recursive=False):
    "Actual implementation on get_leafs."
    return _get_relations_by_id('tail', organisation_id, recursive)

def get_stems_by_id(organisation_id, recursive=False):
    "Actual implementation on get_leafs."
    return _get_relations_by_id('head', organisation_id, recursive)

def get_stems(organisation, recursive=False):
    "Return instances that have organisation as (one of their) leafs."
    tmp = lookups()
    array = get_stems_by_id(tmp['id'][organisation], recursive)
    return _id_to_name_recursive(array)

def get_leafs(organisation, recursive=False):
    "Return instances that have organisation as (one of their) stems."
    tmp = lookups()
    array = get_leafs_by_id(tmp['id'][organisation], recursive)
    return _id_to_name_recursive(array)

def flatten(container, returns=None, prefix=None):
    "Flattens the container."
    if prefix is None:
        prefix = list()
    else:
        prefix = prefix[::]

    if returns is None:
        returns = list()

    if isinstance(container, dict):
        for key, value in container.items():
            returns.append(prefix+[key])
            prefix.append(key)
            flatten(value, returns, prefix)

    elif isinstance(container, (list, tuple, set)):
        for item in container:
            flatten(item, returns, prefix)
    else:
        returns.append(prefix + [container])

    return returns



"""
These are the models at the core of the functionality.
"""
import uuid
from django.db import models
from django.contrib.auth.models import User
from . import abstracts

# pylint: disable=no-member

class Organisation(abstracts.Abstract):
    "Organisations"
    label = models.CharField(max_length=128, unique=True)


class Member(abstracts.Abstract):
    "Defines Membership of organisations."
    user = models.ForeignKey(User, related_name='association_member')
    organisation = models.ForeignKey(Organisation, related_name='member',
                                     blank=True)
    uid_verifier = models.UUIDField(unique=True, default=uuid.uuid4, null=True)
    has_accepted = models.BooleanField(default=False)
    org_managers = models.BooleanField(default=False)
    hrm_managers = models.BooleanField(default=False)

    def __str__(self):
        return str(self.organisation) + ' : ' + str(self.user)


class Association(abstracts.Abstract):
    "Associations"
    _txt_related = 'association_organisation'
    organisation = models.ForeignKey(Organisation, related_name=_txt_related)

    _txt_related = 'association_associations'
    associations = models.ManyToManyField(Organisation, blank=True,
                                          related_name=_txt_related)

    def __str__(self):
        return str(self.organisation) + ' : '  +  str(self.associations.all())



"""
Application Init
"""
import os
from django.apps import AppConfig as _APPCFG
from . import __info__

_ = os.path.abspath(__file__)
_ = os.path.dirname(_)
_ = os.path.split(_)[1]
__info__.LABELS['path'] = _

def _ready(self):
    from . import signals
    return _APPCFG.ready(self)

_ = globals()

_[__info__.LABELS['class']] = \
    type(__info__. LABELS['class'], (_APPCFG,),
         {'ready':_ready, 'name':__info__.LABELS['name'],
          'verbose_name':__info__.LABELS['verbose_name']})

_['default_app_config'] = __info__.LABELS['path'] \
                          + '.' + __info__.LABELS['class']



# -*- coding: utf-8 -*-
# pylint: disable=invalid-name, missing-docstring, line-too-long
# Generated by Django 1.9.6 on 2016-05-20 14:15
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Association',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dts_insert', models.DateTimeField(auto_now_add=True)),
                ('dts_update', models.DateTimeField(blank=True, null=True)),
                ('dts_delete', models.DateTimeField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Member',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dts_insert', models.DateTimeField(auto_now_add=True)),
                ('dts_update', models.DateTimeField(blank=True, null=True)),
                ('dts_delete', models.DateTimeField(blank=True, null=True)),
                ('uid_verifier', models.UUIDField(default=uuid.uuid4, null=True, unique=True)),
                ('has_accepted', models.BooleanField(default=False)),
                ('org_managers', models.BooleanField(default=False)),
                ('hrm_managers', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Organisation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dts_insert', models.DateTimeField(auto_now_add=True)),
                ('dts_update', models.DateTimeField(blank=True, null=True)),
                ('dts_delete', models.DateTimeField(blank=True, null=True)),
                ('label', models.CharField(max_length=128, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='member',
            name='organisation',
            field=models.ForeignKey(blank=True, on_delete=django.db.models.deletion.CASCADE, related_name='member', to='django_associations.Organisation'),
        ),
        migrations.AddField(
            model_name='member',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='association_member', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='association',
            name='associations',
            field=models.ManyToManyField(blank=True, related_name='association_associations', to='django_associations.Organisation'),
        ),
        migrations.AddField(
            model_name='association',
            name='organisation',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='association_organisation', to='django_associations.Organisation'),
        ),
    ]



"""
PyPi Setup file.
"""
# pylint: disable=no-name-in-module, import-error
from setuptools import setup, find_packages
VERSION = '0.1.4'
BASE_URL = 'https://github.com/blancltd/django-associations'
DOWNLOAD = 'https://github.com/blancltd/django-associations/archive/master.zip'

DEPENDENCIES = [
   'django',
   'django-integrator>=0.1.0',
]

CLASSIFIERS = [
   'Programming Language :: Python :: 3',
   'Environment :: Web Environment',
   'Framework :: Django',
   'Framework :: Django :: 1.9',
   'License :: OSI Approved :: BSD License',
   'Operating System :: OS Independent',
]

setup(
   name = 'Django-Org-Associations',
   packages = find_packages(),
   include_package_data=True,
   version = VERSION,
   description = 'Django Organisation/Member associations',
   author = 'Blanc Limited',
   author_email = 'studio@blanc.tld.uk',
   maintainer = 'Martin P. Hellwig',
   maintainer_email = 'martin.hellwig@gmail.com',
   platforms=['any'],
   url = BASE_URL,
   download_url = DOWNLOAD,
   keywords = ['django'],
   license = 'BSD',
   install_requires = DEPENDENCIES,
   classifiers = CLASSIFIERS,
)

